                page    63,132
;   ISR64.ASM
;******************************************************************************
;   ISR64 is an interrupt handler interface to RadioDRV using interrupt 64h.
;   ISR64 updates the range_rate, bf_str (base frequency string), and
;   upd_command (update command) global variables in RadioDRV.
;
;   OrbitDRV, RadioPOP, and subsequent executions of RadioDRV call this module.
;
;   This module was assembled using Borland's Turbo C++ rel. 1.0
;
;   Input:   ah = 3   update range rate velocity
;            ds:dx =  address of velocity as a 64 bit IEEE double
;                     precision floating-point
;
;            ah = 16  slew freq   -10 Hz
;               = 17   "     "    +10 Hz
;               = 18   "     "   -100 Hz
;               = 19   "     "   +100 Hz
;               = 20   "     "  -1000 Hz
;               = 21   "     "  +1000 Hz
;               = 22   Zero offset
;               = 23   Toggle update
;               = 24   Toggle band displayed
;               = 25   Update displayed-band base frequency
;               = 28   Turn display on
;               = 29   Turn display off
;               = 30   Toggle one-shot freq updating
;               = 31   Toggle through band combinations
;               = 32   Toggle inverted doppler for non-displayed band
;               = 33   Return segment into which program is loaded              ; 02-26-93
;               = 40   All other key  presses
;
;            For ah = 25, 26, & 27,
;                ds:si =  address of new base frequency
;
;   Output:  ax  = 0  action completed     (except when ah = 3)                 ; 02-26-93
;            ax != 0  action not completed (except when ah = 3)                 ; 02-26-93
;                                                                               
;******************************************************************************
                IDEAL
                MODEL   TINY,C                                                  ; 02-26-93

                DATASEG
                EXTRN   upd_command:BYTE
                EXTRN   bf_str:BYTE:11
                EXTRN   rr_in:QWORD                     ; range rate            ; 02-26-93
                EXTRN   rr_flag:BYTE                    ; update sentinel       ; 02-26-93
                EXTRN   isTSRflag:WORD                  ; now TSR flag          ; 02-26-93

                CODESEG
                PUBLIC  isr64
                ASSUME  ds:NOTHING                                              ; 02-26-93
                ASSUME  es:NOTHING                                              ; 02-26-93
                ASSUME  ss:NOTHING                                              ; 02-26-93

PROC            isr64   FAR
                jmp     SkipSignature
                ORG     0Ah                             ; offset for OrbitDRV
                db      'RadioDRV','Release 1.3a, December 6, 1994',0
cx_reg          dw      ?                               ; register save area    ; 02-26-93
si_reg          dw      ?                                                       ; 02-26-93
di_reg          dw      ?                                                       ; 02-26-93
es_reg          dw      ?                                                       ; 02-26-93


SkipSignature:
                sti                                     ; enable interrupts
                mov     [cs:isTSRflag],1                ; say we're TSR         ; 02-26-93
range_rate_chg:
                cmp     ah,3                            ; range rate change?
                jne     freq_chg                        ; no, jump
                mov     [cx_reg],cx                     ; save registers        ; 02-26-93
                mov     [si_reg],si                     ; assume nothing        ; 02-26-93
                mov     [di_reg],di                     ;   about callers       ; 02-26-93
                mov     [es_reg],es                     ;   stack space         ; 02-26-93

                mov     cx,cs                           ; put our seg in es     ; 02-26-93
                mov     es,cx                                                   ; 02-26-93
                mov     [cs:rr_flag],1                  ; set update sentinel   ; 02-26-93
                mov     si,dx                           ; put ds:dx in ds:si
                lea     di,[cs:rr_in]                   ; range_rate in es:di   ; 02-26-93
                mov     cx,4                            ; 4 words to move
                cld                                     ; move upward
                rep     movsw                           ; move it
                mov     [cs:rr_flag],0                  ; clear sentinel        ; 02-26-93

                mov     es,[WORD PTR es_reg]            ; restore registers     ; 02-26-93
                mov     di,[WORD PTR di_reg]                                    ; 02-26-93
                mov     si,[WORD PTR si_reg]                                    ; 02-26-93
                mov     cx,[WORD PTR si_reg]                                    ; 02-26-93
                jmp     gohome
freq_chg:
                cmp     ah,25                           ; base freq change?
                jne     segrequest                      ; no, jump
                mov     cx,cs                           ; put our seg in es     ; 02-26-93
                mov     es,cx                                                   ; 02-26-93
                xor     cx,cx                           ; new freq at ds:si
                mov     cl,al                           ; get character count
                lea     di,[cs:bf_str]                  ; get string offset     ; 02-26-93
                cld                                     ; move upward
                rep     movsb                           ; move freq to es:di
                jmp     others
segrequest:
                cmp     ah,33                           ; request for SEG?      ; 02-26-93
                jne     others                          ; no, jump              ; 02-26-93
                mov     bx,cs                           ; return SEG in BX      ; 02-26-93
                mov     ax,0                            ; indicate completed    ; 02-26-93
                jmp     gohome                                                  ; 02-26-93
others:
                mov     [cs:upd_command],ah             ; set update command    ; 02-26-93
gohome:
                iret
ENDP

xerror_ax       dw      ?                               ;storage array for      ; 02-26-93
xerror_bx       dw      ?                               ;  extended error       ; 02-26-93
xerror_cx       dw      ?                               ;  information          ; 02-26-93
xerror_dx       dw      ?                                                       ; 02-26-93
xerror_si       dw      ?                                                       ; 02-26-93
xerror_di       dw      ?                                                       ; 02-26-93
xerror_ds       dw      ?                                                       ; 02-26-93
xerror_es       dw      ?                                                       ; 02-26-93
                dw      3 dup (0)                                               ; 02-26-93
ENDP
                                                                                ; 02-26-93
;****************************************************************************** ; 02-26-93
; xerror_save : Save extended error information available from DOS.             ; 02-26-93
;****************************************************************************** ; 02-26-93
                PUBLIC  xerror_save                                             ; 02-26-93
PROC            xerror_save NEAR                                                ; 02-26-93
;                cmp    [cs:dosversion],030Ah           ; DOS 3.10 or later?    ; 02-26-93
;                jb      end_save                       ; no, then jump         ; 02-26-93
                push    ds                              ; save DS               ; 02-26-93
                mov     ah,59h                          ; get extended error    ; 02-26-93
                sub     bx,bx                           ;   information         ; 02-26-93
                int     21h                                                     ; 02-26-93
                mov     [xerror_ds],ds                  ; value returned in DS  ; 02-26-93
                pop     ds                              ; restore DS            ; 02-26-93
                mov     [xerror_ax],ax                  ; save return values    ; 02-26-93
                mov     [xerror_bx],bx                  ;   in XERROR array     ; 02-26-93
                mov     [xerror_cx],cx                                          ; 02-26-93
                mov     [xerror_dx],dx                                          ; 02-26-93
                mov     [xerror_si],si                                          ; 02-26-93
                mov     [xerror_di],di                                          ; 02-26-93
                mov     [xerror_es],es                                          ; 02-26-93
end_save:                                                                       ; 02-26-93
                ret                                                             ; 02-26-93
ENDP                                                                            ; 02-26-93
                                                                                ; 02-26-93
;****************************************************************************** ; 02-26-93
; xerror_restore : Restore extended error information                           ; 02-26-93
;****************************************************************************** ; 02-26-93
                PUBLIC xerror_restore                                           ; 02-26-93
PROC            xerror_restore NEAR                                             ; 02-26-93
;                cmp    [cs:dosversion],030Ah           ; DOS 3.10 or later?    ; 02-26-93
;                jb      end_restore                    ; no, then jump         ; 02-26-93
                mov     ax,cs                           ; make sure DS points   ; 02-26-93
                mov     ds,ax                           ;  to this segment      ; 02-26-93
                mov     ax,5D0Ah                        ; restore information   ; 02-26-93
                mov     dx,OFFSET xerror_ax             ;   with DOS function 5D; 02-26-93
                int     21h                                                     ; 02-26-93
end_restore:                                                                    ; 02-26-93
                ret                                                             ; 02-26-93
ENDP                                                                            ; 02-26-93
                                                                                ; 02-26-93
myint02h        dd      ?                               ; my intr 02h vector    ; 02-26-93
myint34h        dd      ?                               ; my intr 34h vector    ; 02-26-93
myint35h        dd      ?                               ; my intr 35h vector    ; 02-26-93
myint36h        dd      ?                               ; my intr 36h vector    ; 02-26-93
myint37h        dd      ?                               ; my intr 37h vector    ; 02-26-93
myint38h        dd      ?                               ; my intr 38h vector    ; 02-26-93
myint39h        dd      ?                               ; my intr 39h vector    ; 02-26-93
myint3Ah        dd      ?                               ; my intr 3Ah vector    ; 02-26-93
myint3Bh        dd      ?                               ; my intr 3Bh vector    ; 02-26-93
myint3Ch        dd      ?                               ; my intr 3Ch vector    ; 02-26-93
myint3Dh        dd      ?                               ; my intr 3Dh vector    ; 02-26-93
myint3Eh        dd      ?                               ; my intr 3Eh vector    ; 02-26-93
                                                                                ; 02-26-93
oldint02h       dd      ?                               ; old intr 02h vector   ; 02-26-93
oldint34h       dd      ?                               ; old intr 34h vector   ; 02-26-93
oldint35h       dd      ?                               ; old intr 35h vector   ; 02-26-93
oldint36h       dd      ?                               ; old intr 36h vector   ; 02-26-93
oldint37h       dd      ?                               ; old intr 37h vector   ; 02-26-93
oldint38h       dd      ?                               ; old intr 38h vector   ; 02-26-93
oldint39h       dd      ?                               ; old intr 39h vector   ; 02-26-93
oldint3Ah       dd      ?                               ; old intr 3Ah vector   ; 02-26-93
oldint3Bh       dd      ?                               ; old intr 3Bh vector   ; 02-26-93
oldint3Ch       dd      ?                               ; old intr 3Ch vector   ; 02-26-93
oldint3Dh       dd      ?                               ; old intr 3Dh vector   ; 02-26-93
oldint3Eh       dd      ?                               ; old intr 3Eh vector   ; 02-26-93
                                                                                ; 02-26-93
;****************************************************************************** ; 02-26-93
; save_my_fpvec : Save my floating-point emulation-library vectors              ; 02-26-93
;****************************************************************************** ; 02-26-93
                ASSUME  ds:NOTHING                                              ; 02-26-93
                ASSUME  es:NOTHING                                              ; 02-26-93
                ASSUME  ss:NOTHING                                              ; 02-26-93
                PUBLIC save_my_fpvec                                            ; 02-26-93
                                                                                ; 02-26-93
PROC            save_my_fpvec NEAR                                              ; 02-26-93
                                                                                ; 02-26-93
                push    cx                                                      ; 02-26-93
                push    ds                                                      ; 02-26-93
                push    es                                                      ; 02-26-93
                push    di                                                      ; 02-26-93
                push    si                                                      ; 02-26-93
                mov     cx,0                            ; segment 0 in DS
                mov     ds,cx
                push    cs                              ; our segment in ES
                pop     es
                mov     si,8                            ; intr 02h off in SI
                lea     di,[myint02h]                   ; save area off in DI
                mov     cx,2                            ; words to move
                cld                                     ; move upward
                rep     movsw                           ; move it
                mov     si,0D0h                         ; intr 34h off in SI
                lea     di,[myint34h]                   ; save area off in DI
                mov     cx,22                           ; words to move
                rep     movsw                           ; move it

                pop     si
                pop     di
                pop     es
                pop     ds
                pop     cx
                ret
ENDP
                                                                                ; 02-26-93
;****************************************************************************** ; 02-26-93
; swap_fpvec : SWAP current floating-point emulation-library vectors            ; 02-26-93
;              and install our vectors: 02, 34 - 3E                             ; 02-26-93
;****************************************************************************** ; 02-26-93
                ASSUME  ds:NOTHING                                              ; 02-26-93
                ASSUME  es:NOTHING                                              ; 02-26-93
                ASSUME  ss:NOTHING                                              ; 02-26-93
                PUBLIC swap_fpvec                                               ; 02-26-93
                                                                                ; 02-26-93
PROC            swap_fpvec NEAR                                                 ; 02-26-93
                                                                                ; 02-26-93
                push    cx                              ; save registers        ; 02-26-93
                push    ds                                                      ; 02-26-93
                push    es                                                      ; 02-26-93
                push    di                                                      ; 02-26-93
                push    si                                                      ; 02-26-93

                ; save current interrupt vectors

                mov     cx,0                            ; source - DS:SI
                mov     ds,cx                           ;   segment 0 in DS
                push    cs                              ; destination - ES:DI
                pop     es                              ;   our segment in ES

                mov     si,8                            ; intr 02h off in SI
                lea     di,[oldint02h]                  ; save area off in DI
                mov     cx,2                            ; words to move
                cld                                     ; move upward
                rep     movsw                           ; move it

                mov     si,0D0h                         ; intr 34h off in SI
                lea     di,[oldint34h]                  ; save area off in DI
                mov     cx,22                           ; words to move
                rep     movsw                           ; move it

                ; install our interreupt vectors

                push    cs                              ; source - DS:SI
                pop     ds                              ;   our segment in DS
                mov     cx,0                            ; destination - ES:DI
                mov     es,cx                           ;   segment 0 in ES

                lea     si,[myint02h]                   ; save area off in SI
                mov     di,8                            ; intr 02h off in DI
                mov     cx,2                            ; words to move
                cld                                     ; move upward
                cli                                     ; disable interrupts
                rep     movsw                           ; move it

                lea     si,[myint34h]                   ; save area off in SI
                mov     di,0D0h                         ; intr 34h off in DI
                mov     cx,22                           ; words to move
                rep     movsw                           ; move it

                sti                                     ; enable interrupts
                pop     si                              ; restore registers
                pop     di
                pop     es
                pop     ds
                pop     cx
                ret
ENDP

;****************************************************************************** ; 02-26-93
; restore_old_fpvec : Install old floating-point emulation-library vectors      ; 02-26-93
;****************************************************************************** ; 02-26-93
                ASSUME  ds:NOTHING                                              ; 02-26-93
                ASSUME  es:NOTHING                                              ; 02-26-93
                ASSUME  ss:NOTHING                                              ; 02-26-93
                PUBLIC  restore_old_fpvec                                       ; 02-26-93
                                                                                ; 02-26-93
PROC            restore_old_fpvec NEAR                                          ; 02-26-93
                                                                                ; 02-26-93
                push    cx                              ; save registers        ; 02-26-93
                push    ds                                                      ; 02-26-93
                push    es                                                      ; 02-26-93
                push    di                                                      ; 02-26-93
                push    si                                                      ; 02-26-93

                ; restore the vectors we swapped out earlier

                push    cs                              ; source - DS:SI
                pop     ds                              ;   our segment in DS
                mov     cx,0                            ; destination - ES:DI
                mov     es,cx                           ;   segment 0 in ES

                lea     si,[oldint02h]                  ; save area off in SI
                mov     di,8                            ; intr 02h off in DI
                mov     cx,2                            ; words to move
                cld                                     ; move upward
                cli                                     ; disable interrupts
                rep     movsw                           ; move it

                lea     si,[oldint34h]                  ; save area off in SI
                mov     di,0D0h                         ; intr 34h off in DI
                mov     cx,22                           ; words to move
                rep     movsw                           ; move it

                sti                                     ; enable interrupts
                pop     si                              ; restore registers
                pop     di
                pop     es
                pop     ds
                pop     cx
                ret
ENDP

state87         db      100 DUP (?)                     ; 94 needed plus spare  ; 02-26-93
                                                                                ; 02-26-93
;****************************************************************************** ; 02-26-93
; fnsave : save co-processor state                                              ; 02-26-93
;****************************************************************************** ; 02-26-93
                ASSUME  ds:NOTHING                                              ; 02-26-93
                ASSUME  es:NOTHING                                              ; 02-26-93
                ASSUME  ss:NOTHING                                              ; 02-26-93
                PUBLIC  fnsave                                                  ; 02-26-93
                                                                                ; 02-26-93
PROC            fnsave NEAR                                                     ; 02-26-93
                                                                                ; 02-26-93
                cli                                     ; disable interrupts    ; 02-26-93
                fwait                                   ; wait till its idle    ; 02-26-93
                fnsave  [cs:state87]                    ; save the state        ; 02-26-93
                fwait                                   ; wait till its idle    ; 02-26-93
                sti                                     ; enable interrupts     ; 02-26-93
                ret                                                             ; 02-26-93
ENDP
                                                                                ; 02-26-93
;****************************************************************************** ; 02-26-93
; frstor : restore co-processor state                                           ; 02-26-93
;****************************************************************************** ; 02-26-93
                ASSUME  ds:NOTHING                                              ; 02-26-93
                ASSUME  es:NOTHING                                              ; 02-26-93
                ASSUME  ss:NOTHING                                              ; 02-26-93
                PUBLIC  frstor                                                  ; 02-26-93
                                                                                ; 02-26-93
PROC            frstor NEAR                                                     ; 02-26-93
                                                                                ; 02-26-93
;;;;;;;         cli          shouldn't need it here     ; disable interrupts    ; 02-26-93
                fwait                                   ; wait till its idle    ; 02-26-93
                frstor  [cs:state87]                    ; restore the state     ; 02-26-93
                fwait                                   ; wait till its idle    ; 02-26-93
;;;;;;;         sti          shouldn't need it here     ; enable interrupts     ; 02-26-93
                ret                                                             ; 02-26-93

ENDP

                ENDS
                END
