#include "mbbiosi.h"                                                            // 12-06-94

// -------------------------------------------------------------------------
//   comparams : Initialize the comm port
//               Returns modem satus register
// -------------------------------------------------------------------------
char comparams(struct comtype *com)
{
   static struct REGPACK reg;

   reg.r_dx = com->port - 1;
   switch ( (int) com->speed ) {                                               // 12-06-94
      case  110 : reg.r_ax = 0x00; break;
      case  150 : reg.r_ax = 0x20; break;
      case  300 : reg.r_ax = 0x40; break;
      case  600 : reg.r_ax = 0x60; break;
      case 1200 : reg.r_ax = 0x80; break;
      case 2400 : reg.r_ax = 0xA0; break;
      case 4800 : reg.r_ax = 0xC0; break;
      case 9600 : reg.r_ax = 0xE0; break;
   }

   switch (com->parity) {
      case 'N' : break;
      case 'n' : break;
      case 'O' : reg.r_ax |= 0x08; break;
      case 'o' : reg.r_ax |= 0x08; break;
      case 'E' : reg.r_ax |= 0x18; break;
      case 'e' : reg.r_ax |= 0x18; break;
   }

   switch (com->stopbits) {
      case 1 : break;
      case 2 : reg.r_ax |= 0x04; break;
   }

   switch (com->databits) {
      case 7 : reg.r_ax |= 0x02; break;
      case 8 : reg.r_ax |= 0x03; break;
   }
   intr (0x14, &reg);
   return reg.r_ax;                    // return modem status register
}

// -------------------------------------------------------------------------
//   comwritech : Write character to the comm port using MBBIOS
//                Returns modem satus register
// -------------------------------------------------------------------------
char comwritech(int port, char c)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_ax = 0x0100 + c;
   intr (0x14, &reg);
   return reg.r_ax;                              // return MSR
}

// -------------------------------------------------------------------------
//   comreadch : Read character from the com port using MBBIOS
//               Returns composite line satus registers
// -------------------------------------------------------------------------
char comreadch(int port, char *c)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_ax = 0x0200;
   intr (0x14, &reg);
   *c = reg.r_ax;                                // Get AL
   return reg.r_ax >> 8;
}

// -------------------------------------------------------------------------
//   comstatus: Get the modem status register
// -------------------------------------------------------------------------
char comstatus(int port)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_ax = 0x0300;
   intr (0x14, &reg);
   return reg.r_ax;
}

// -------------------------------------------------------------------------
//   ckMBBIOS : Check if MBBIOS is loaded for a particular port
//              Returns 0 if yes, 1 if no
// -------------------------------------------------------------------------
int ckMBBIOS(int port)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_ax = 0x0400;
   intr (0x14, &reg);
   if (reg.r_ax == 0xAA55)
      return 0;
   else
      return 1;
}

// -------------------------------------------------------------------------
//   dropRTS
// -------------------------------------------------------------------------
void dropRTS(int port)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_ax = 0x0500;
   intr (0x14, &reg);
}

// -------------------------------------------------------------------------
//   raiseRTS
// -------------------------------------------------------------------------
void raiseRTS(int port)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_ax = 0x0600;
   intr (0x14, &reg);
}

// -------------------------------------------------------------------------
//   sendbrk  : Send break
// -------------------------------------------------------------------------
void sendbrk(int port)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_ax = 0x0700;
   intr (0x14, &reg);
}

// -------------------------------------------------------------------------
//   dataready : See if there is a character in the buffer without
//               removine it.  Sets "c" to the received character.
//               Returns 1 if there is, otherwise return 0
// -------------------------------------------------------------------------
int dataready(int port, char *c)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_ax = 0x0800;
   intr (0x14, &reg);
   *c = reg.r_ax;                                // Get AL
   return (reg.r_ax >> 8) & RDR;
}

// -------------------------------------------------------------------------
//   optMBBIOS: Set/get Options
//              Returns old options byte
// -------------------------------------------------------------------------
char optMBBIOS(int port, char opt)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_ax = 0x0900 + opt;
   intr (0x14, &reg);
   return reg.r_ax;
}

// -------------------------------------------------------------------------
//   comwrite : Write buffer to the comm port using MBBIOS
//              returns unsent character count
// -------------------------------------------------------------------------
int comwrite(int port, char buffer[], int count)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_es = FP_SEG (buffer);
   reg.r_di = FP_OFF (buffer);
   reg.r_ax = 0x0A00;
   reg.r_cx = count;
   intr (0x14, &reg);
   return (reg.r_cx);                  // return unsent character count
}

// -------------------------------------------------------------------------
//   comread : Read buffer from the comm port
//             returns character count read
// -------------------------------------------------------------------------
int comread(int port, char buffer[], int count)
{
   static struct REGPACK reg;

   reg.r_dx = port - 1;
   reg.r_es = FP_SEG (buffer);
   reg.r_di = FP_OFF (buffer);
   reg.r_ax = 0x0B00;
   reg.r_cx = count;
   intr (0x14, &reg);
   return reg.r_cx;                    // return count read
}

