// **************************************************************************
//
//   Radio Driver - TSR - Release 1.3 - February 26, 1993
//
//   Jesse Buckwalter, NZ3F, Annapolis, MD, developed this program to
//   control the frequency of the ICOM IC-970.  It should work with other
//   ICOM products that have a computer interface as long as the correct
//   parameters are used.  No guarantees!  With some changes, you could
//   adapt the logic to rigs of other manufacturers.  This program is
//   freeware.  Do with it whatever you wish.  Hopefully, some of you use
//   it as a model for developing radio-driver programs that you will share.
//
//   NOTE:  You must compile this program using the tine memory model.
//          Segment references are based on the tiny model.
//
//   Change history:
//
//   12-05-92:  Correction of addressing error, provided by Wayne Roth, WA2N
//   12-09-92:  Added IF frequency parameters
//   12-10-92:  Added frequency field length parameters
//   02-26-93:  Emiminated use of OrbitDRV stack in isr64
//              Added vector swap for floating-point emulation library vectors
//              Added saving/restoring of co-processor state
//              Added check for InDos and DOS critical error flags
//              Added DOS Idle-loop isr28
//              Switched ISRs to use internal stack (isr1C and isr28)
//              Changed to tiny memory model
//              Added save and restore of DOS extended-error information
//              Cleaned up other minor coding errors
//
// **************************************************************************

#include <dos.h>
#include <bios.h>                                                               // 02-26-93
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>                      // tolower()
#include <string.h>
#include <float.h>                                                              // 02-26-93
#include "mbbiosi.h"                    // MBBIOS interface functions

#define NO      0
#define YES     1
#define MAIN    0
#define SUB     1
#define BOTH    2
#define NEITHER 3

extern unsigned        _heaplen = 1024;                                         // 02-26-93
extern unsigned        _stklen = 1024;                                          // 02-26-93
//            assembler coded functions                                         // 02-26-93

extern "C"                                                                      // 12-06-94
{                                                                               // 12-06-94
   void interrupt isr64( ... );              // RadioDRV interface interrupt    // 12-06-94

   void   xerror_save( void );               // save and restore extended       // 12-06-94
   void   xerror_restore( void );            //   DOS error info                // 12-06-94
   void   save_my_fpvec( void );             // save my fp vectors              // 12-06-94
   void   swap_fpvec( void );                // hook current vectors            // 12-06-94
   void   restore_old_fpvec( void );         // restore current vectors         // 12-06-94
   void   fnsave( void );                    // save co=processor state         // 12-06-94
   void   frstor( void );                    // restore co-processor state      // 12-06-94
};


// ICOM CI-V Interface messages 0x00 and 0x07

struct {                                         // set freq, no response
   char prefix[2];
   char radio_addr;                                                             // 12-05-92
   char computer_addr;                                                          // 12-05-92
   char cmd_nr;
   char freq[5];
   char suffix;
} msg0 = {0xFE, 0xFE, 0x00, 0xE0, 0x00, 0, 0, 0, 0, 0, 0xFD};

static struct {                                  // select main/sub band
   char prefix[2];
   char radio_addr;                                                             // 12-05-92
   char computer_addr;                                                          // 12-05-92
   char cmd_nr;
   char sub_cmd_nr;
   char suffix;
} msg7 = {0xFE, 0xFE, 0x00, 0xE0, 0x07, 0, 0xFD};


typedef void interrupt (*CastIsr) (...);         // define the type cast        // 12-06-94
CastIsr oldvec1C;                                // old clock-tick vector       // 12-06-94
CastIsr oldvec28;                                // old idle vector             // 12-06-94

char          upd_command        = 0;            // set by RadioPOP & RadioRDV
char far*     cmd_ptr;                           // pointer to update command   // 02-26-93
char          bf_str[11];                        // set by RadioPOP & RadioDRV
double        rr_in              = 0;            // range rate set by isr64     // 02-26-93
double        range_rate         = 0;            // range rate                  // 02-26-93
char          rr_flag            = 0;            // range_rate sentinel         // 02-26-93
unsigned      bflag              = NEITHER;      // band combos to update
unsigned far* bflg_ptr;                          // ptr to band combos flag     // 12-06-94
unsigned      uflag              = NO;           // update freqs flag
unsigned far* uflg_ptr;                          // ptr to update freqs flag    // 12-06-94
unsigned      u_bnd              = NEITHER;      // current band to update
unsigned      d_bnd              = SUB;          // current band to display
unsigned      invert             = NO;           // invert offset
char          bandfunct          = 0xD0;         // command 07 sub-command
unsigned long base_freq[2]       = {145000000L, 437000000L};
unsigned long freq[2]            = {145000000L, 437000000L};
unsigned long old_freq[2]        = {145000000L, 437000000L};
unsigned long far *bf_ptr;                       // pointer to base freq array  // 02-26-93
int           flen[2]            = {5, 5};       // freq length in bytes        // 12-10-92
unsigned long IF[2]              = {0, 0};       // intermediate frequency      // 12-09-92
unsigned long IF_offset[2]       = {0, 0};       // IF offset from basefreq     // 12-09-92
long          offset[2]          = {0, 0};
int           doppler[2]         = {0, 0};       // current doppler             // 02-26-93
int           lastdoppler[2]     = {0, 0};       // doppler at last change      // 02-26-93
unsigned      interval           = 10;
struct        comtype com        = {1, 9600, 'N', 8, 1};
char          rcvr_addr          = 46;           // ICOM IC-970 address
char          xmtr_addr          = 46;           // ICOM IC-970 address
char          BCD_Ch[100];
int           attr               = 0x4E00;       // screen attribute (color)
unsigned      (far *screen)[80];
char          bnd_flag_ch[]      = "MS&-";
unsigned      dflag              = NO;           // RadioPOP active flag
unsigned      osflag             = NO;           // one-shot freq update flag
unsigned long tcount             = 0;            // total clock-ticks counter   // 02-26-93
int           icount             = 0;            // interval ticks counter      // 02-26-93

int           isTSRflag          = 0;            // program now resident flag   // 02-26-93
int           requestflag        = 0;            // RadioDRV request flag       // 02-26-93
int           busyflag           = 0;            // RadioDRV busy flag          // 02-26-93

char far *    inDOSflag;                         // pointer to InDos flag       // 02-26-93
char far *    critERRflag;                       // pointer to CritErr flag     // 02-26-93
char          numcoflag          = 0;            // numeric co-processor flag   // 02-26-93
              char state87[94];                  // saved co-processor state    // 02-26-93

unsigned      mystack;                           // top of internal stack       // 02-26-93
int           ss_register;                       // register save area          // 02-26-93
int           sp_register;                                                      // 02-26-93

// --------------------------------------------------------------------------
//   itoBCD : Convert unsigned long to an array containing BCD
//            itoBCD[0] returns number of data bytes starting in itoBCD[1]
// --------------------------------------------------------------------------
char *itoBCD(unsigned long x)
{
   unsigned long  x1;
   static char *  s  = "12345";                                                 // 02-26-93
   char           i;

   x1 = x;
   i = 0;
   do
   {
      i++;
      s[i] = BCD_Ch[(short) (x1 % 100)];
      x1 = x1 / 100;
   } while (x1 != 0);
   s[0] = i;
   return s;
}

// --------------------------------------------------------------------------
//   init_BCD_matrix : Build integer-to-BCD conversion matrix
// --------------------------------------------------------------------------
void init_BCD_matrix(void)
{
   int m, n;

   for (m = 0; m < 10; m++)
      for (n = 0; n < 10; n++)
         BCD_Ch[10*m + n] = 16*m + n;
}

// --------------------------------------------------------------------------
//   justr : Justify string to right and pad left with designated character
// --------------------------------------------------------------------------
void justr(char fstr[], int flen, char ch)
{
   int i, j, diff;

   diff = flen - strlen(fstr);
   for (i = flen; i >= 0; i--)
   {
      j = i - diff;
      if (j >= 0)
         fstr[i] = fstr[j];
      else
         fstr[i] = ch;
   }
   return;
}

// --------------------------------------------------------------------------
// fmtfreq : format frequency string with periods
// --------------------------------------------------------------------------
void fmtfreq(char *c)
{
   c[0] = c[2];
   c[1] = c[3];
   c[2] = c[4];
   c[3] = c[5];
   c[4] = '.';
   c[5] = c[6];
   c[6] = c[7];
   c[7] = c[8];
   c[8] = '.';
}

// --------------------------------------------------------------------------   // 02-26-93
// calc_doppler:  Calculate doppler for a band                                  // 02-26-93
//                This is the only function in RadioDRV that manipulates        // 02-26-93
//                floating-point numbers.  Because the Borland emulation        // 02-26-93
//                library is not re-entrant, special steps must be takes to     // 02-26-93
//                preserve the integrity of foreground tasks that use the       // 02-26-93
//                same library interrupts: 34, 35, 36, 37, 38, 39, 3A, 3B       // 02-26-93
//                3C, 3D, & 3E.  When we were installed, we hooked these        // 02-26-93
//                vectors.  The foreground task also hooks these vectors.       // 02-26-93
//                Unfortunately, the library routines update their own data     // 02-26-93
//                in the stack segment.  There are ways to maintain integrity   // 02-26-93
//                by preserving these data, if we knew where was the stack      // 02-26-93
//                for the foreground task.  We don't.                           // 02-26-93
//                                                                              // 02-26-93
//                The workaround is to swap in our vectors.  We make sure the   // 02-26-93
//                SS and SP registers are ours so our emulation data is         // 02-26-93
//                updated.  We restore the foreground vectors before we         // 02-26-93
//                leave.                                                        // 02-26-93
//                                                                              // 02-26-93
//                If a numeric co-processor is installed, we must save the      // 02-26-93
//                state of the 80x87 chip.  After saving the state, an implicit // 02-26-93
//                FINIT is executed to mask all exceptions.  The code is        // 02-26-93
//                based on Borland's TI722.                                     // 02-26-93
// --------------------------------------------------------------------------   // 02-26-93
int calc_doppler( unsigned band )                                               // 02-26-93
{                                                                               // 02-26-93
   static int d;                                                                // 02-26-93
                                                                                // 02-26-93
   if (numcoflag)                                // if numeric co-processor     // 02-26-93
      fnsave();                                  //    save its state           // 02-26-93
   else                                          // otherwise                   // 02-26-93
      swap_fpvec();                              //    replace emu vectors      // 02-26-93
                                                                                // 02-26-93
   if (!rr_flag)                                 // get range_rate when it's    // 02-26-93
      range_rate = rr_in;                        //   NOT being updated         // 02-26-93
   d = - range_rate * base_freq[band];           // do our calculations         // 02-26-93
                                                                                // 02-26-93
   if (numcoflag)                                // if numeric co-processor     // 02-26-93
      frstor();                                  //    restore its state        // 02-26-93
   else                                          // otherwise                   // 02-26-93
      restore_old_fpvec();                       //    restore emu vectors      // 02-26-93
   return d;                                     // return the value            // 02-26-93
}                                                                               // 02-26-93

// --------------------------------------------------------------------------
// calc:  Calculate doppler compensated frequency for a band
// --------------------------------------------------------------------------
void calc( unsigned band )
{                                                                               // 02-26-93
   static int           changedop[2];            // change in doppler           // 02-26-93
   static int           int_dopp[2]   =  {0, 0}; // change in doppler/interval  // 02-26-93
   static int           int_tick[2]   = {36,36}; // change in ticks/interval    // 02-26-93
   static int           startdop[2]   =  {0, 0}; // starting doppler per cycle
   static unsigned long startcnt[2]   =  {0, 0}; // start count for cycle

   // Calculated new doppler.  Interpolate between OrbitDRV updates to
   // the range-rate if there has been no change.

   doppler[band] = calc_doppler(band);

   // Set values for the initial calculation or after a base-frequecy change.
   // Disregard "tcount" which is approx. equal to "interval" at the start.

   if (lastdoppler[band] == 0)
      lastdoppler[band] = startdop[band] = doppler[band];

   changedop[band]   = doppler[band] - lastdoppler[band];
   lastdoppler[band] = doppler[band];

   if (changedop[band] == 0)

      // Increment doppler by the prorated change in doppler for the last       // 02-26-93
      // interval of actual doppler change                                      // 02-26-93

      doppler[band] += (int) ((long) (tcount - startcnt[band]) *                // 02-26-93
                              int_dopp[band] / int_tick[band]);                 // 02-26-93
   else
   {
      // Calculate interval doppler and interval time in ticks since the last   // 02-26-93
      // doppler change.                                                        // 02-26-93

      int_tick[band] = (int) (tcount - startcnt[band]);                         // 02-26-93
      int_dopp[band] = doppler[band] - startdop[band];                          // 02-26-93

      startcnt[band] = tcount;                   // reset counter
      startdop[band] = doppler[band];
   }

   if (band == MAIN)
      freq[band] = base_freq[band] - doppler[band]; // adjusted for doppler
   else
      freq[band] = base_freq[band] + doppler[band]; // adjusted for doppler
}

// --------------------------------------------------------------------------
//    display: Display the data values on the screen
// --------------------------------------------------------------------------
void display(unsigned band)
{
   static char f_str[13];                        // alphanum freq string
   static int  sl;                               // string length
   int         i;

   if (!dflag)                                   // RadioPOP activated?
      return;
   ltoa(offset[band], f_str, 10);
   justr(f_str, 12, ' ');
   fmtfreq(f_str);
   sl = strlen(f_str);
   for (i = 0; i < sl; i++)
      screen[4][60+i] = f_str[i] + attr;         // display offset

   ultoa(base_freq[band], f_str, 10);
   justr(f_str, 12, ' ');
   fmtfreq(f_str);
   sl = strlen(f_str);
   for (i = 0; i < sl; i++)
      screen[5][60+i] = f_str[i] + attr;         // display base freq

   itoa(doppler[band], f_str, 10);
   justr(f_str, 12, ' ');
   fmtfreq(f_str);
   sl = strlen(f_str);
   for (i = 0; i < sl; i++)
      screen[6][60+i] = f_str[i] + attr;         // display doppler

   ultoa(freq[band], f_str, 10);
   justr(f_str, 12, ' ');
   fmtfreq(f_str);
   sl = strlen(f_str);
   for (i = 0; i < sl; i++)                      // display freq
      screen[7][60+i] = f_str[i] + attr;

   screen[4][58] = bnd_flag_ch[band] + attr;     // display band symbol
   if (uflag)
      screen[7][73] = 'U' + attr;                // display update symbol
   else
      if (osflag)
         screen[7][73] = 'O' + attr;             // display one-shot symbol
      else
         screen[7][73] = '-' + attr;             // display no symbol

   screen[7][74] = bnd_flag_ch[bflag] + attr;    // display band combo symbol
   if (invert)
      screen[7][75] = 'I' + attr;                // display invert symbol
   else
      screen[7][75] = '-' + attr;
}

// --------------------------------------------------------------------------
//   update: Update designated band of the transceiver with the current freq
// --------------------------------------------------------------------------
void update(unsigned band)
{
   static char   *BCDptr;                                                       // 02-26-93
   unsigned long  newfreq;                                                      // 12-09-92
   static char   *zero = "\x0\x0\x0\x0\x0";                                     // 12-09-92

   // Select transmitter address (main) or                                      // 12-05-92
   // receiver address (sub)                                                    // 12-05-92
                                                                                // 12-05-92
   switch(band)                                                                 // 12-05-92
   {                                                                            // 12-05-92
      case MAIN : msg0.radio_addr = xmtr_addr; break;                           // 12-05-92
      case SUB  : msg0.radio_addr = rcvr_addr;                                  // 12-05-92
   }                                                                            // 12-05-92

   // Correct for IF output (default offset is zero)                            // 12-09-92
                                                                                // 12-09-92
   newfreq = freq[band] - IF_offset[band];                                      // 12-09-92

   // Update transmitter/receiver

   BCDptr = itoBCD(newfreq);                     // get BCD freq                // 12-09-92
   memcpy(msg0.freq, zero, 5);                   // init data area              // 12-09-92
   memcpy(msg0.freq, &BCDptr[1], BCDptr[0]);     // insert BCD freq
   if (flen[band] == 4)                          // handle 4 byte freq field    // 12-10-92
   {                                                                            // 12-10-92
      ((char *) msg0.freq)[4] = msg0.suffix;     // more up suffix              // 12-10-92
      comwrite(com.port, (char *) &msg0, sizeof(msg0) - 1);                     // 12-10-92
   }                                                                            // 12-10-92
   else                                                                         // 12-10-92
      comwrite(com.port, (char *) &msg0, sizeof(msg0));
}

// --------------------------------------------------------------------------   // 02-26-93
//   radiodrv_work: main memory-resident function                               // 02-26-93
//                                                                              // 02-26-93
//   This routine calculates a new frequency and updates the ICOM               // 02-26-93
//   receiver with the new value.   This function also checks the upd_command   // 02-26-93
//   for changes.  isr64 changes upd_command after interrupts from OrbitDRV,    // 02-26-93
//   RadioPOP, and subsequent executions of RadioDRV.                           // 02-26-93
//                                                                              // 02-26-93
//   Both the isr1C and isr28 handlers can call this function, depending on     // 02-26-93
//   the state of Dos.                                                          // 02-26-93
// --------------------------------------------------------------------------   // 02-26-93
void radiodrv_work( void )                                                      // 02-26-93
{
   static int           update_state  = 0;       // update state variable       // 02-26-93
   static int           df;                      // offset freq increment
   static unsigned      one_shot      = NO;      // one_shot this time
   char                *endptr;                                                 // 02-26-93

   xerror_save();                                // save extended err info      // 02-26-93
   if (upd_command != 0)               // update command is set by RadioPOP or
   {                                   //   main program by values passed to
      df = 0;                          //   isr64 in register AH at interrupt
      switch (upd_command)
      {
         case 16: df =   -10; break;             // get offset increment
         case 17: df =   +10; break;
         case 18: df =  -100; break;
         case 19: df =  +100; break;
         case 20: df = -1000; break;
         case 21: df = +1000; break;
         case 22:                                // zero offset
            if (bflag == MAIN || bflag == BOTH)
            {
               base_freq[MAIN] -= offset[MAIN];
               offset[MAIN]     = 0;
               freq[MAIN]       = base_freq[MAIN] - calc_doppler(MAIN);         // 02-26-93
               old_freq[MAIN]   = freq[MAIN];
            }
            if (bflag == SUB || bflag == BOTH)
            {
               base_freq[SUB] -= offset[SUB];
               offset[SUB]     = 0;
               freq[SUB]       = base_freq[SUB] + calc_doppler(SUB);            // 02-26-93
               old_freq[SUB]   = freq[SUB];
            }
            break;
         case 23:                                // toggle frequency update
            uflag  = !uflag;
            osflag = NO;
            break;
         case 24:                                // toggle band displayed
            d_bnd = !d_bnd;
            break;
         case 25:                                // set bf of band displayed
            base_freq[d_bnd] = strtoul(bf_str, &endptr, 10); // via interrupt   // 02-26-93
            if (d_bnd == MAIN)                                                  // 02-26-93
               freq[MAIN]    = base_freq[MAIN] - calc_doppler(MAIN);            // 02-26-93
            else                                                                // 02-26-93
               freq[SUB]     = base_freq[SUB]  + calc_doppler(SUB);             // 02-26-93
            old_freq[d_bnd]  = freq[d_bnd];
            offset[d_bnd]    = 0;
            doppler[d_bnd]   = lastdoppler[d_bnd] = 0;
            break;
         case 26:                                // set main-band base freq
            freq[MAIN]       = base_freq[MAIN] - calc_doppler(MAIN);            // 02-26-93
            old_freq[MAIN]   = freq[MAIN];
            offset[MAIN]     = 0;
            doppler[MAIN]    = lastdoppler[MAIN] = 0;                           // 02-26-93
            break;
         case 27:                                // set sub-band base freq
            freq[SUB]        = base_freq[SUB] + calc_doppler(SUB);              // 02-26-93
            old_freq[SUB]    = freq[SUB];
            doppler[SUB]     = lastdoppler[SUB] = 0;                            // 02-26-93
            offset[SUB]      = 0;
            break;
         case 28:                                // turn display on
            dflag = YES;
            break;
         case 29:                                // turn display off
            dflag = NO;
            break;
         case 30:                                // toggle one-shot freq update
            osflag = !osflag;
            uflag  = NO;
            break;
         case 31:                                // toggle through band combos
            bflag = (bflag + 1) % 4;
            break;
         case 32:                                // toggle offset inversion
            invert = !invert;
            break;                                                              // 12-06-94
         case 40:                                // on-shot update this pass
            if (!osflag)                         // ignore misc key-pounding
               break;
            one_shot = YES;
            break;
      }
      if (df)                                    // test for offset increment
      {
         if (bflag == d_bnd || bflag == BOTH)
         {
            base_freq[d_bnd] += df;              // offset base frequency
            offset[d_bnd]    += df;              // save cummulative offset
         }
         if (bflag == !d_bnd || bflag == BOTH)
         {
            if (invert)                          // invert increment for the
            {                                    // band that is not displayed
               base_freq[!d_bnd] -= df;
               offset[!d_bnd]    -= df;
            }
            else
            {
               base_freq[!d_bnd] += df;
               offset[!d_bnd]    += df;
            }
         }
      }
      upd_command = 0;
      icount = interval;                         // do updates ASAP, although
   }                                             // that skews interpolation

   // Update freq every "interval" clock ticks.  Bands are changed, and
   // frequencies are updated for sub and main bands on successive ticks.
   // interval +0, +1, +2, +3                                                   // 02-26-93

   if (icount >= interval)                        // updating is done based on  // 02-26-93
      switch (update_state)                                                     // 02-26-93
      {
         case 0:                                 // 1st tick in cycle
            update_state = 1;                                                   // 02-26-93
//            if (u_bnd == MAIN ||                 // already MAIN?
//               (bflag != MAIN && bflag != BOTH)) // not interested in MAIN?
            if (bflag != MAIN && bflag != BOTH)  // not interested in MAIN?
               break;                            // yes, no band change
            if (uflag || one_shot)               // updating?
            {
               u_bnd = MAIN;
               msg7.sub_cmd_nr = bandfunct + u_bnd; // yes, select band
               msg7.radio_addr = rcvr_addr;                                     // 12-05-92
               comwrite(com.port, (char *) &msg7, sizeof(msg7));
            }
            break;
         case 1:                                 // 2nd tick in cycle
            update_state = 2;                                                   // 02-26-93
            calc(MAIN);
            display(d_bnd);
            if ((uflag || one_shot) && (bflag == MAIN || bflag == BOTH))
               update(MAIN);
            break;
         case 2:                                 // 3rd tick in cycle
            update_state = 3;                                                   // 02-26-93
//            if (u_bnd == SUB ||                  // already SUB?
//               (bflag != SUB && bflag != BOTH))  // not interested in SUB?
            if (bflag != SUB && bflag != BOTH)   // not interested in SUB?
               break;                            // yes, no band change
            if (uflag || one_shot)               // updating?
            {
               u_bnd = SUB;
               msg7.sub_cmd_nr = bandfunct + u_bnd; // yes, select band
               comwrite(com.port, (char *) &msg7, sizeof(msg7));
            }
            break;
         case 3:                                 // 4th tick in cycle
            update_state = 0;                                                   // 02-26-93
            calc(SUB);
            display(d_bnd);
            if ((uflag || one_shot) && (bflag == SUB || bflag == BOTH))
               update(SUB);
            one_shot = NO;                       // reset
            icount = 0;
            break;
      }
   xerror_restore();                             // restore extended err info   // 02-26-93
}

// --------------------------------------------------------------------------   // 02-26-93
//   isr28 : Interrupt Service Routine 0x28 - Dos Idle Interrupt                // 02-26-93
// --------------------------------------------------------------------------   // 02-26-93
void interrupt isr28( ... )                                                    // 02-26-93
{                                                                               // 02-26-93
   oldvec28();                                   // call old handler            // 02-26-93
   enable();                                     // enable interrupts           // 02-26-93
                                                                                // 02-26-93
   if (!requestflag ||                           // if RadioDRV not requested   // 02-26-93
       busyflag     ||                           // if RadioDRV busy            // 02-26-93
       *critERRflag    )                         // if Dos critical error       // 02-26-93
      return;                                    //   then exit                 // 02-26-93
   busyflag = 1;                                 // say RadioDRV is busy        // 02-26-93
   requestflag = 0;                              // clear request flag          // 02-26-93
                                                                                // 02-26-93
   // Load stack registers to use internal stack rather than the callers.       // 02-26-93
                                                                                // 02-26-93
   disable();                                    // disable interrupts          // 02-26-93
   ss_register = _SS;                            // save callers stack seg      // 02-26-93
   sp_register = _SP;                            // save callers stack ptr      // 02-26-93
   _SS = _CS;                                    // set ss to our code seg      // 02-26-93
   _SP = mystack;                                // set sp to internal stack    // 02-26-93
   enable();                                     // enable interrupts           // 02-26-93
                                                                                // 02-26-93
   radiodrv_work();                              // do the real work            // 02-26-93
                                                                                // 02-26-93
   // Restore the callers stack registers before exiting                        // 02-26-93
                                                                                // 02-26-93
   disable();                                    // disable interrupts          // 02-26-93
   _SS = ss_register;                            // restore stack seg           // 02-26-93
   _SP = sp_register;                            // restore stack pointer       // 02-26-93
   enable();                                     // enable interrupts           // 02-26-93
   busyflag = 0;                                 // say RadioDRV isn't busy     // 02-26-93
}                                                                               // 02-26-93
                                                                                // 02-26-93
// --------------------------------------------------------------------------   // 02-26-93
//   isr1C : Interrupt Service Routine 0x1C - Clock Tick Interrupt              // 02-26-93
//                                                                              // 02-26-93
//   The BIOS clock tick interrupt calls this routines 18.2 times per second.   // 02-26-93
//   If Dos is busy, then it sets the request flag and exits.  The Dos          // 02-26-93
//   idle-interrupt handler, isr28, then handles the request.                   // 02-26-93
// --------------------------------------------------------------------------   // 02-26-93
void interrupt isr1C( ... )                                                    // 02-26-93
{                                                                               // 02-26-93
   oldvec1C();                                   // call old handler            // 02-26-93
   if (!isTSRflag)                               // if program not yet          // 02-26-93
      return;                                    //   resident then exit        // 02-26-93
                                                                                // 02-26-93
   icount++;                                     // increment tick counters     // 02-26-93
   tcount++;                                                                    // 02-26-93
                                                                                // 02-26-93
   if (*inDOSflag   ||                           // if Dos is busy              // 02-26-93
       *critERRflag ||                           // if Dos critical error       // 02-26-93
       busyflag       )                          // if RadioDRV busy            // 02-26-93
   {                                                                            // 02-26-93
      requestflag = 1;                           // set RadioDRV request flag   // 02-26-93
      return;                                                                   // 02-26-93
   }                                                                            // 02-26-93
   busyflag = 1;                                 // say RadioDRV is busy        // 02-26-93
                                                                                // 02-26-93
   // Load stack registers to use internal stack rather than the callers.       // 02-26-93
                                                                                // 02-26-93
   disable();                                    // disable interrupts          // 02-26-93
   ss_register = _SS;                            // save callers stack seg      // 02-26-93
   sp_register = _SP;                            // save callers stack ptr      // 02-26-93
   _SS = _CS;                                    // set ss to our code seg      // 02-26-93
   _SP = mystack;                                // set sp to internal stack    // 02-26-93
   enable();                                     // enable interrupts           // 02-26-93
                                                                                // 02-26-93
   radiodrv_work();                              // do the real work            // 02-26-93
                                                                                // 02-26-93
   // Restore the callers stack registers before exiting                        // 02-26-93
                                                                                // 02-26-93
   disable();                                    // disable interrupts          // 02-26-93
   _SS = ss_register;                            // restore stack seg           // 02-26-93
   _SP = sp_register;                            // restore stack pointer       // 02-26-93
   enable();                                     // enable interrupts           // 02-26-93
   busyflag = 0;                                 // say RadioDRV isn't busy     // 02-26-93
}                                                                               // 02-26-93
                                                                                // 02-26-93
// --------------------------------------------------------------------------
//   setupcom : Initialize the serial port
// --------------------------------------------------------------------------
void setupcom (void)
{
   char i_buf[] = "######";

   if (ckMBBIOS( com.port ) != 0)                // MBBIOS loaded?
   {
      printf( "MBBIOS not loaded for com port %i\n", com.port );
      exit( 1 );
   }
   optMBBIOS( com.port, TRB );                   // options: xmit buffering,
                                                 //   no handshaking
   comparams( &com );                            // set up com port
   while (comread( com.port, i_buf, 7 ) != 0);   // empty input buffer

   // Test comm port
   comwrite( com.port, "@@@@@@", 6 );
   delay( 150 );
   if (comread( com.port, i_buf, 6 ) == 0 ||
       strncmp( i_buf, "@@@@@@", 6 ) != 0)
   {
      printf("Serial-port link not active.\nRadioDRV installation halted.\n");
      exit( 1 );
   }
}

// --------------------------------------------------------------------------
//   options1(c):  processes a parameter string "c"
// --------------------------------------------------------------------------
//
//     format: [-,/][aaaa][=bbbb]                           default
//
//     [A1=<xx>]                     = address -> band 1    (/a1=46)
//     [A2=<xx>]                     = address -> band 2    (/a2=46)
//     [A3=<xx>]                     = address -> computer  (/a3=224)
//     [B0 | B1 | B2 | B&            = band to update       (/b0)
//     [DAta=<x>]                    = comm port databits   (/da=8)
//     [D1 | D2]                     = display band         (/d2)
//     [F1=<xxxxxxxxx>]              = freq -> band 1       (/f1=145000000)
//     [F2=<xxxxxxxxx>]              = freq -> band 2       (/f2=437000000)
//     [I1=<xxxxxxxxx>]              = IF -> band 1         (/i1=0)             // 12-09-92
//     [I2=<xxxxxxxxx>]              = IF -> band 2         (/i2=0)             // 12-09-92
//     [INvert | NInvert]            = invert offset        (/ni)
//     [L1=<x>]                      = band 1 freq length   (/l1=5);            // 12-10-92
//     [L2=<x>]                      = band 2 freq length   (/l2=5);            // 12-10-92
//     [MOno | COlor]                = monochrome mode      (/co)
//     [PArity=<c>]                  = comm port parity     (/pa=N)
//     [POrt=<x>]                    = comm port number     (/po=1)
//     [SPeed=<xxxx>]                = comm port speed      (/sp=9600)
//     [VFos | MS-MainSub]           = band function        (/ms)
//     [STop=<x>]                    = comm port stopbits   (/st=1)
//     [UPdate | NUpdate]            = turn on/off updating (/nu)
//     [UInterval=<xx>]              = update interval      (/ui=10)            // 02-17-93
//
// --------------------------------------------------------------------------
int options1( char *c )
{
   char        p[20], *param, *arg, *endptr;
   int         i , l;

   if (c[0] != '/' && c[0] != '-')                                              // 02-26-93
   {
      printf("Error in parameter: %s\n", c);
      return 1;
   }
   l = 19 < strlen( c ) ? 19 : strlen( c );                                     // 12-06-94
   strncpy( p, c, l );
   p[l] = '\0';
   for (i = 0; i < l; i++)
      p[i] = tolower(p[i]);
   param = strtok(p, "/-= \r\n");
   arg = strtok(NULL, " \r\n");
   switch ((param[0]))
   {
      case 'a':
         switch (param[1])
         {
            case '1':                            // main band (xmit) address
               xmtr_addr = (char) atoi(arg);                                    // 12-05-92
               break;
            case '2':                            // sub band (rcvr) address
               rcvr_addr = (char) atoi(arg);                                    // 12-05-92
               break;
            case '3':                            // computer address            // 12-05-92
               msg0.computer_addr = (char) atoi(arg);                           // 12-05-92
               msg7.computer_addr = (char) atoi(arg);                           // 12-05-92
               break;                                                           // 12-05-92
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'b':
         switch (param[1])
         {
            case '&':                            // band combo flag = BOTH
               bflag = BOTH;
               break;
            case '1':                            // band combo flag = MAIN
               bflag = MAIN;
               break;
            case '0':                            // band combo flag = NEITHER
               bflag = NEITHER;
               break;
            case '2':                            // band combo flag = SUB
               bflag = SUB;
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'c':
         switch (param[1])
         {
            case'o':                             // color mode, set attribute
               attr = 0x4E00;
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'd':
         switch (param[1])
         {
            case 'a':                            // comm port data bits
               com.databits = atoi(arg);
               break;
            case '1':                            // display flag = MAIN
               d_bnd = MAIN;
               break;
            case '2':                            // display flag = SUB
               d_bnd = SUB;
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'f':
         switch (param[1])
         {
            case '1':                            // main-band base freq
               old_freq[MAIN] = freq[MAIN] = base_freq[MAIN] = atol(arg);
               break;
            case '2':                            // sub-band base freq
               old_freq[SUB] = freq[SUB] = base_freq[SUB] = atol(arg);
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'i':
         switch (param[1])
         {
            case '1':                            // main IF frequency           // 12-09-92
               IF[MAIN] = strtoul(arg, &endptr, 10);                            // 02-26-93
               break;                                                           // 12-09-92
            case '2':                            // sub IF frequency            // 12-09-92
               IF[SUB] = strtoul(arg, &endptr, 10);                             // 02-26-93
               break;                                                           // 12-09-92
            case 'n':                            // turn on offset inverting
               invert = YES;
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'l':                                                                 // 12-10-92
         switch (param[1])                                                      // 12-10-92
         {                                                                      // 12-10-92
            case '1':                            // main freq length            // 12-10-92
               flen[MAIN]=atoi(arg);                                            // 12-10-92
               break;                                                           // 12-10-92
            case '2':                            // sub freq length             // 12-10-92
               flen[SUB]=atoi(arg);                                             // 12-10-92
               break;                                                           // 12-10-92
            default:                                                            // 12-10-92
               printf("Error in parameter: %s\n", c);                           // 12-10-92
               return 1;                                                        // 12-10-92
         }                                                                      // 12-10-92
         break;                                                                 // 12-10-92
      case 'm':
         switch (param[1])
         {
            case 'o':                            // monochrome attributes
               attr = 0x0700;
               break;
            case 's':                            // main-sub band function
               bandfunct = 0xD0;
               strcpy(bnd_flag_ch, "MS&-");      // use MS instead of AB
               break;                            
            default:                             // default
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'n':
         switch (param[1])
         {
            case 'i':                            // turn off offset inverting
               invert = NO;
               break;
            case 'u':                            // turn off freq updating
               uflag = NO;
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'p':
         switch (param[1])
         {
            case 'a':                            // comm port parity
               com.parity = arg[0];
               break;
            case 'o':                            // comm port number
               com.port = (char) atoi(arg);
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 's':
         switch (param[1])
         {
            case 'p':                            // comm port speed
               com.speed = atoi(arg);
               break;
            case 't':                            // comm port stop bits
               com.stopbits = atoi(arg);
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'u':
         switch (param[1])
         {
            case 'i':                            // update interval (freq)
               interval = atoi(arg);
               break;
            case 'p':                            // turn on freq-update flag
               uflag = YES;
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'v':
         switch (param[1])
         {
            case 'f':                            // vfo A / B band function
               bandfunct = 0x00;
               strcpy(bnd_flag_ch, "AB&-");      // use AB instead of MS
               break;
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      default:
         printf("Error in parameter: %s\n", c);
         return 1;
   }
   return 0;
}

// --------------------------------------------------------------------------
//   options2(c):  Process a parameter string "c" when RadioDRV is already
//                 installed.  Change values by interrupt 0x64.
// --------------------------------------------------------------------------
int options2( char *c )
{
   char            p[20], *param, *arg, *endptr;                                // 02-26-93
   int             i;
   int             l;

   if (c[0] != '/' && c[0] != '-')                                              // 02-17-93
   {
      printf("Error in parameter: %s\n", c);
      return 1;
   }
   l = 19 < strlen( c ) ? 19 : strlen( c );                                     // 12-06-94
   strncpy( p, c, l );                                                          // 02-17-93
   p[l] = 0;
   for (i = 0; i < l; i++)
      p[i] = tolower(p[i]);
   param = strtok(p, "/-= \r\n");
   arg = strtok(NULL, " \r\n");
   if (strlen(arg) > 10)                         // 10 digits max               // 02-26-93
      param[1] = 0;                              // create param error          // 02-26-93
   switch ((param[0]))
   {
      case 'b':                                                                 // 12-06-94
         switch (param[1])                                                      // 12-06-94
         {                                                                      // 12-06-94
            case '&':                            // band combo flag = BOTH      // 12-06-94
               *bflg_ptr = BOTH;                                                // 12-06-94
               printf( "Bands to update changed to BOTH\n" );                   // 12-06-94
               break;                                                           // 12-06-94
            case '1':                            // band combo flag = MAIN      // 12-06-94
               *bflg_ptr = MAIN;                                                // 12-06-94
               printf( "Bands to update changed to MAIN\n" );                   // 12-06-94
               break;                                                           // 12-06-94
            case '0':                            // band combo flag = NEITHER   // 12-06-94
               *bflg_ptr = NEITHER;                                             // 12-06-94
               printf( "Bands to update changed to NEITHER\n" );                // 12-06-94
               break;                                                           // 12-06-94
            case '2':                            // band combo flag = SUB       // 12-06-94
               *bflg_ptr = SUB;                                                 // 12-06-94
               printf( "Bands to update changed to SUB\n" );                    // 12-06-94
               break;                                                           // 12-06-94
            default:                                                            // 12-06-94
               printf("Error in parameter: %s\n", c);                           // 12-06-94
               return 1;                                                        // 12-06-94
         }                                                                      // 12-06-94
         break;                                                                 // 12-06-94
      case 'f':
         switch (param[1])
         {
            case '1':                            // main-band base freq
               bf_ptr[MAIN] = strtoul(arg, &endptr, 10);                        // 02-26-93
               *cmd_ptr = 26;                    // indicate it has changed;    // 02-26-93
               for (i = 0; i < 10; i ++)         // wait for change to take     // 02-26-93
               {                                                                // 02-26-93
                  if (!*cmd_ptr)                                                // 02-26-93
                     break;                                                     // 02-26-93
                  delay( 100 );                                                 // 02-26-93
               }                                                                // 02-26-93
               if (!*cmd_ptr)                                                   // 02-26-93
                  printf("Base frequency 1 is now %s\n", arg);                  // 02-26-93
               else                                                             // 02-26-93
                  printf("RadioDRV not responding to frequency change");        // 02-26-93
               break;                                                           // 02-26-93
            case '2':                            // sub-band base freq          // 02-26-93
               bf_ptr[SUB] = strtoul(arg, &endptr, 10);                         // 02-26-93
               *cmd_ptr = 27;                    // indicate it has changed;    // 02-26-93
               for (i = 0; i < 10; i ++)         // wait for change to take     // 02-26-93
               {                                                                // 02-26-93
                  if (!*cmd_ptr)                                                // 02-26-93
                     break;                                                     // 02-26-93
                  delay( 100 );                                                 // 02-26-93
               }                                                                // 02-26-93
               if (!*cmd_ptr)                                                   // 02-26-93
                  printf("Base frequency 2 is now %s\n", arg);                  // 02-26-93
               else                                                             // 02-26-93
                  printf("RadioDRV not responding to frequency change");        // 02-26-93
               break;                                                           // 02-26-93
            default:
               printf("Error in parameter: %s\n", c);
               return 1;
         }
         break;
      case 'n':                                                                 // 12-06-94
         switch (param[1])                                                      // 12-06-94
         {                                                                      // 12-06-94
            case 'u':                            // turn off freq updating      // 12-06-94
               *uflg_ptr = NO;                                                  // 12-06-94
               printf( "Frequency updating turned OFF\n" );                     // 12-06-94
               break;                                                           // 12-06-94
            default:                                                            // 12-06-94
               printf("Error in parameter: %s\n", c);                           // 12-06-94
               return 1;                                                        // 12-06-94
         }                                                                      // 12-06-94
         break;                                                                 // 12-06-94
      case 'u':                                                                 // 12-06-94
         switch (param[1])                                                      // 12-06-94
         {                                                                      // 12-06-94
            case 'p':                            // turn on freq-update flag    // 12-06-94
               *uflg_ptr = YES;                                                 // 12-06-94
               printf( "Frequency updating turned ON\n" );                      // 12-06-94
               break;                                                           // 12-06-94
            default:                                                            // 12-06-94
               printf("Error in parameter: %s\n", c);                           // 12-06-94
               return 1;                                                        // 12-06-94
         }                                                                      // 12-06-94
         break;                                                                 // 12-06-94
      default:
         printf("Error in parameter: %s\n", c);
         return 1;
   }
   return 0;
}

// --------------------------------------------------------------------------
//   main program root
// --------------------------------------------------------------------------
void main( int argc, char *argv[] )
{
   #define                 CO_PROCESSOR_MASK 0x0002
   char                    sig1[] = "RotorDRV+OrbitDRV";                        // 12-06-94
   char                    sig2[] = "RadioDRV";                                 // 12-06-94
   char far*               sig;                  // signature pointer
   int                     i;
   int                     err = 0;
   char far*               mem;
   unsigned                pgm_size;                                            // 02-26-93
   union REGS              xreg;                                                // 02-26-93
   struct SREGS            sreg;                                                // 02-26-93

   // The following variables are used in the debugginhg code below
   // char buf[1000], ch;
   // int count;

   printf("RadioDRV 1.3a");                      // announce the version        // 12-06-94
   if ((((_version & 0x00FF) << 8) + (_version >> 8)) < 0x030A)                 // 02-26-93
   {                                                                            // 02-26-93
      printf(" requires DOS 3.1 or higher\n");                                  // 02-26-93
      return;                                                                   // 02-26-93
   }                                                                            // 02-26-93

   sig = (char far*) getvect( 0x63 ) + 0x0A;     // see if OrbitDRV is loaded
   i = 0;                                        //   by looking for signature
   while ( sig[i] == sig1[i] ) i++;              // can't use memicmp with far
   if (i < strlen( sig1 ))                       //   pointers                  // 12-06-94
   {
      printf("\nOrbitDRV not installed; RadioDRV installation halted.\n");
      exit(1);
   }

   // isr64 provides the "RadioDRV" signature and the interrupt 0x64
   // interface to RadioDRV.
                                                 
   sig = (char far*) getvect( 0x64 ) + 0x0A;     // see if RadioDRV is loaded
   i = 0;                                        //   by looking for signature
   while ( sig[i] == sig2[i] ) i++;
   if (i >= strlen( sig2 ))                                                     // 12-06-94
   {
      printf(" already installed\n");            // RadioDRV installed
      for (i = 0; i < 10; i++)                   // loop in case ISR is busy    // 02-26-93
      {                                          // get resident data segment   // 02-26-93
         xreg.h.ah = 33;                         // cmd value (33) in AH        // 02-26-93
         int86(0x64, &xreg, &xreg);                                             // 02-26-93
         if (xreg.x.ax == 0)                     // zero means change took      // 02-26-93
            break;                                                              // 02-26-93
         delay(50);                              // wait a bit                  // 02-26-93
      }                                                                         // 02-26-93
      if (xreg.x.ax != 0)                                                       // 02-26-93
      {                                                                         // 02-26-93
         printf("Memory resident RadioDRV not responding.\n");                  // 02-26-93
         return;                                                                // 02-26-93
      }                                                                         // 02-26-93
                                                 // segment returned in BX      // 12-06-94
      bf_ptr = (unsigned long far*) MK_FP(xreg.x.bx, FP_OFF(base_freq));        // 12-06-94
      cmd_ptr = (char far*) MK_FP(xreg.x.bx, FP_OFF(&upd_command));             // 12-06-94            // 02-26-93
      bflg_ptr = (unsigned far*) MK_FP(xreg.x.bx, FP_OFF(&bflag));              // 12-06-94
      uflg_ptr = (unsigned far*) MK_FP(xreg.x.bx, FP_OFF(&uflag));              // 12-06-94
      for (i = 1; i < argc; i++)                 // get command line options
         options2(argv[i]);                                                     // 02-26-93
      return;
   }
   if (biosequip() & CO_PROCESSOR_MASK)          // RadioDRV not installed      // 02-26-93
   {                                             // check for co-processor      // 02-26-93
      numcoflag = 1;                                                            // 02-26-93
      printf(" installation started, co-processor will be used.\n");            // 02-26-93
   }                                                                            // 02-26-93
   else                                                                         // 02-26-93
      printf(" installation started, no co-processor detected.\n");             // 02-26-93

   for (i = 1; i < argc; i++)                    // get command line options
      err += options1(argv[i]);
   if (err != 0)
   {
      printf("RadioDRV installation halted\n");
      return;
   }
   if (IF[MAIN] != 0)                                                           // 12-09-92
      IF_offset[MAIN] = base_freq[MAIN] - IF[MAIN]; // get offset               // 12-09-92
   if (IF[SUB] != 0)                                                            // 12-09-92
      IF_offset[SUB] = base_freq[SUB] - IF[SUB];    // get offset               // 12-09-92
   setupcom();                                   // initialize comm port
   init_BCD_matrix();                            // init conversion matrix
   mem = (char far*) MK_FP(0x0000, 0x0449);      // check video adapter         // 12-06-94
   if (*mem == 7)
   {
       screen = (unsigned (far*)[]) MK_FP( 0xB000, 0 ); // monochrome mode        // 12-06-94
       attr = 0x0700;
   }
   else
       screen = (unsigned (far*)[]) MK_FP( 0xB800, 0 ); // color mode             // 12-06-94
   oldvec1C = getvect(0x1C);                     // get clock-tick vector       // 02-26-93
   oldvec28 = getvect(0x28);                     // get idle vector             // 02-26-93
   save_my_fpvec();                              // save my fp vectors          // 02-26-93
   _control87( MCW_EM, MCW_EM);                  // mask fp exceptions          // 02-26-93

   xreg.x.ax = 0x3400;                           // get InDos flag addr         // 02-26-93
   intdosx(&xreg, &xreg, &sreg);                                                // 02-26-93
   inDOSflag = (char far *) MK_FP(sreg.es, xreg.x.bx);                          // 02-26-93
   critERRflag = inDOSflag - 1;                  // get CritErr flag addr       // 02-26-93
   printf("RadioDRV installation completed; ");
   printf("base freqs are %lu, %lu\n", base_freq[MAIN], base_freq[SUB]);

   // Prepare to Terminate and Stay Resident.  First disable all interrupts     // 02-26-93
   // so that we can set the "isTSRflag" without risk of a timer tick           // 02-26-93
   // occuring before we have "gone tsr."                                       // 02-26-93
                                                                                // 02-26-93
   disable();
   setvect(0x1C, isr1C);                         // install RadioDRV ISRs
   setvect(0x28, isr28);                                                        // 02-26-93
   setvect(0x64, isr64);                                                        // 02-26-93

   // _psp is 0x100 bytes long and precedes the code segment.  Because          // 02-26-93
   // the tiny model is used, all offsets are relative to the same segment.     // 02-26-93
   // Therefore the program size in paragraphs can be calculated by adding      // 02-26-93
   // the psp size to stack pointer, dividing by 16, and rounding up.           // 02-26-93
                                                                                // 02-26-93
   pgm_size = (0x100 + _SP + 15) / 16;                                          // 02-26-93
   mystack = _SP;                                // save our stack offset       // 02-26-93
   keep(0, pgm_size);                            // terminate and stay resident // 02-26-93
   return;
}



